package com.polopoly.modeshape;

import java.util.Iterator;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.concurrent.locks.ReentrantLock;

import javax.inject.Inject;
import javax.jcr.LoginException;
import javax.jcr.NoSuchWorkspaceException;
import javax.jcr.Node;
import javax.jcr.NodeIterator;
import javax.jcr.Property;
import javax.jcr.PropertyIterator;
import javax.jcr.Repository;
import javax.jcr.RepositoryException;
import javax.jcr.Session;
import javax.jcr.lock.Lock;
import javax.jcr.lock.LockException;
import javax.ws.rs.Consumes;
import javax.ws.rs.GET;
import javax.ws.rs.PUT;
import javax.ws.rs.Path;
import javax.ws.rs.PathParam;
import javax.ws.rs.Produces;

import org.codehaus.jettison.json.JSONArray;
import org.codehaus.jettison.json.JSONException;
import org.codehaus.jettison.json.JSONObject;

@Path("/")
public class ModeshapeService {
    private final Repository _repository;
    
    static AtomicInteger call = new AtomicInteger();
    private static final ReentrantLock pathWriterLock = new ReentrantLock();

    @Inject
    public ModeshapeService(Repository repository)
        throws LoginException, NoSuchWorkspaceException, RepositoryException
    {
        _repository = repository;
        Session session = _repository.login("main");
        try {
            if (!session.getRootNode().isNodeType("mix:lockable")) {
                pathWriterLock.lock();
                try {
                    session.refresh(false);
                    if (!session.getRootNode().isNodeType("mix:lockable")) {
                        session.getRootNode().addMixin("mix:lockable");
                        session.save();
                        System.out.println("Added lockable to root (thread " + Thread.currentThread().getId() + ")");
                    }
                } finally {
                    System.out.println("Root unlocked  (thread " + Thread.currentThread().getId() + ")");
                    pathWriterLock.unlock();
                }
            }
        } finally {
            session.logout();
        }
    }

    @Path("document/{path}")
    @GET
    @Produces("application/json")
    public JSONObject getDocument(@PathParam("path") String path) throws Exception {
        Session session = _repository.login("main");
        Node rootNode = session.getRootNode();
        Node node = rootNode.getNode(path);

        JSONObject json = jsonify(node);

        return json;
    }

    @Path("document/{path}/count")
    @GET
    @Produces("application/json")
    public JSONObject count(@PathParam("path") String path)
        throws Exception
    {
        Session session = _repository.login("main");
        JSONObject json = new JSONObject();
        try {
            Node rootNode = session.getRootNode();
            Node pnode = rootNode.getNode(path);
            NodeIterator nodeIter = pnode.getNodes();
            long size = nodeIter.getSize();
            while (nodeIter.hasNext()) {
                Node inode = nodeIter.nextNode();
                //System.out.println(inode.getPath());
                NodeIterator subNodeIter = inode.getNodes();
                size += subNodeIter.getSize();
                while (subNodeIter.hasNext()) {
                    Node tnode = subNodeIter.nextNode();
                    //System.out.println(tnode.getPath());
                }
            }
            json.put("count", size);
        } finally {
            session.logout();
        }
        return json;
    }

    private Node getTopNode( Session session, String path ) throws Exception {
        if (!session.getRootNode().hasNode(path)) {
            // The node doesn't yet exist, so get a lock and try again ...
            Boolean retry = true;
            while (retry) {
                retry = false;
                Lock lock = null;
                if(!session.getWorkspace().getLockManager().isLocked(session.getRootNode().getPath())) {
                    try {
                        do {
                            Thread.sleep(10);
                            session.refresh(false);
                        } while(session.getWorkspace().getLockManager().isLocked(
                                session.getRootNode().getPath()));

                        // Another thread may have created the node while we waited for the lock,
                        // so look for the node again
                        if (!session.getRootNode().hasNode(path)) {
                            lock = session.getWorkspace().getLockManager().lock(
                                    session.getRootNode().getPath(),
                                    false, true, Long.MAX_VALUE, path);
                            session.getRootNode().addNode(path);
                            System.out.println("Created node at " + path );
                            session.save();
                            System.out.println("Saved new node (thread " + Thread.currentThread().getId() + ") at " + path );
                            long numTopLevelNodes = session.getRootNode().getNodes().getSize();
                            System.out.println("Number of top-level nodes: " + numTopLevelNodes );
                        }
                    } catch (LockException e) {
                        // There is a lock at the root node, lets try again!
                    	e.printStackTrace();
                        retry = true;
                        System.out.println("Lock retry path: (thread " + Thread.currentThread().getId() + ") " + path);
                    }
                    finally {
                        if (null != lock) {
                            System.out.println("Unlock path " + session.getRootNode().getPath());
                            session.getWorkspace().getLockManager().unlock(lock.getNode().getPath());
                        }
                    }
                } else {
                    lock = session.getWorkspace().getLockManager().getLock(
                            session.getRootNode().getPath());
                    // Is the right node for waiting created?
                    if (lock.getLockOwner().equals(path)) {
                        while (!session.getRootNode().hasNode(path)) {
                            System.out.println("Find retry path: (thread " + Thread.currentThread().getId() + ") "+ path);
                            Thread.sleep(10);
                            session.refresh(false);
                        }
                    } else {
                        // No we need a lock!
                        retry = true;
                    }
                }
            }
        }
        return session.getRootNode().getNode(path);
    }

    @Path("document/{path}")
    @PUT
    @Consumes("application/json")
    @Produces("text/plain")
    public String createDocument(@PathParam("path") String path, JSONObject json)
        throws Exception
    {
        Integer incr = call.incrementAndGet();
        long start = System.nanoTime();
        Session session = _repository.login("main");
        try {
            Node rootNode = getTopNode(session,path);
            String nodeName = incr.toString();
            Node numNode = rootNode.addNode(nodeName);
            for (int i = 0;i < 10; i++) {
                Node node = numNode.addNode(String.valueOf(i));
                //session.getRootNode().addNode(path + "/" + incr.toString() + "/" + i);
                node.setPrimaryType("nt:unstructured");
                writeJsonToNode(json, node);
            }
            session.save();
            //System.out.println("Doing PUT " + incr.toString());
        
        } finally {
            session.logout();
        }

        long end = System.nanoTime();
        return String.format("Creating 10 nodes took %dms", (end - start) / 1000000);
    }

    private void writeJsonToNode(JSONObject json, Node node)
        throws Exception {
        Iterator<?> keys = json.keys();
        while (keys.hasNext()) {
            String key = (String) keys.next();
            Object object = json.get(key);
            if (object instanceof JSONObject) {
                Node child = node.addNode(key);
                writeJsonToNode((JSONObject) object, child);
            }
            else if (object instanceof JSONArray) {
                JSONArray array = (JSONArray) object;
                for (int i = 0; i < array.length(); i++) {
                    Object item = array.get(i);
                    if (item instanceof JSONObject) {
                        Node itemNode = node.addNode(key);
                        writeJsonToNode((JSONObject) item, itemNode);
                    }
                    else {
                        // ToDo: create list
                        setProperty(node, key, item);
                    }
                }
            }
            else {
                setProperty(node, key, object);
            }
        }
    }

    private void setProperty(Node node, String key, Object object)
        throws Exception {
        if (object instanceof Long) {
            node.setProperty(key, (Long)object);
        } else if (object instanceof Integer) {
            node.setProperty(key, (Integer)object);
        } else if (object instanceof Double) {
            node.setProperty(key, (Double)object);
        } else if (object instanceof String) {
            node.setProperty(key, (String)object);
        } else if (object instanceof Boolean) {
            node.setProperty(key, (Boolean)object);
        } else {
            throw new IllegalArgumentException();
        }
    }

    private JSONObject jsonify(Node node)
        throws RepositoryException, JSONException
    {
        NodeIterator nodes = node.getNodes();
        JSONObject json = new JSONObject();
        while (nodes.hasNext()) {
            Node next = (Node) nodes.next();
            NodeIterator nodes2 = node.getNodes(next.getName());
            if (nodes2.getSize() > 1) {
                JSONArray array = new JSONArray();
                nodes.skip(nodes2.getSize() - 1);
                while (nodes2.hasNext()) {
                    array.put(jsonify(nodes2.nextNode()));
                }
                json.put(next.getName(), array);
            }
            else {
                json.put(next.getName(), jsonify(next));
            }
        }

        PropertyIterator properties = node.getProperties();
        while (properties.hasNext()) {
            Property property = (Property) properties.next();
            json.put(property.getName(), property.getString());
        }
        return json;
    }
}
